@@ -22,6 +22,10 @@ class _DiscardType:
     """When the Discard value is returned from a transformer callback,
     that node is discarded and won't appear in the parent.
 
+    Note:
+        This feature is disabled when the transformer is provided to Lark
+        using the ``transformer`` keyword (aka Tree-less LALR mode).
+
     Example:
         ::
 
@@ -68,19 +72,24 @@ class _Decoratable:
 
 
 class Transformer(_Decoratable, ABC, Generic[_Leaf_T, _Return_T]):
-    """Transformers visit each node of the tree, and run the appropriate method on it according to the node's data.
+    """Transformers work bottom-up (or depth-first), starting with visiting the leaves and working
+    their way up until ending at the root of the tree.
 
-    Methods are provided by the user via inheritance, and called according to ``tree.data``.
-    The returned value from each method replaces the node in the tree structure.
+    For each node visited, the transformer will call the appropriate method (callbacks), according to the
+    node's ``data``, and use the returned value to replace the node, thereby creating a new tree structure.
 
-    Transformers work bottom-up (or depth-first), starting with the leaves and ending at the root of the tree.
     Transformers can be used to implement map & reduce patterns. Because nodes are reduced from leaf to root,
     at any point the callbacks may assume the children have already been transformed (if applicable).
 
+    If the transformer cannot find a method with the right name, it will instead call ``__default__``, which by
+    default creates a copy of the node.
+
+    To discard a node, return Discard (``lark.visitors.Discard``).
+
     ``Transformer`` can do anything ``Visitor`` can do, but because it reconstructs the tree,
     it is slightly less efficient.
 
-    To discard a node, return Discard (``lark.visitors.Discard``).
+    A transformer without methods essentially performs a non-memoized partial deepcopy.
 
     All these classes implement the transformer interface:
 
@@ -93,7 +102,6 @@ class Transformer(_Decoratable, ABC, Generic[_Leaf_T, _Return_T]):
                                        Setting this to ``False`` is slightly faster. Defaults to ``True``.
                                        (For processing ignored tokens, use the ``lexer_callbacks`` options)
 
-    NOTE: A transformer without methods essentially performs a non-memoized partial deepcopy.
     """
     __visit_tokens__ = True   # For backwards compatibility
 
@@ -150,7 +158,11 @@ class Transformer(_Decoratable, ABC, Generic[_Leaf_T, _Return_T]):
 
     def transform(self, tree: Tree[_Leaf_T]) -> _Return_T:
         "Transform the given tree, and return the final result"
-        return self._transform_tree(tree)
+        res = list(self._transform_children([tree]))
+        if not res:
+            return None     # type: ignore[return-value]
+        assert len(res) == 1
+        return res[0]
 
     def __mul__(
             self: 'Transformer[_Leaf_T, Tree[_Leaf_U]]',
@@ -259,7 +271,7 @@ class TransformerChain(Generic[_Leaf_T, _Return_T]):
         return TransformerChain(*self.transformers + (other,))
 
 
-class Transformer_InPlace(Transformer):
+class Transformer_InPlace(Transformer[_Leaf_T, _Return_T]):
     """Same as Transformer, but non-recursive, and changes the tree in-place instead of returning new instances
 
     Useful for huge trees. Conservative in memory.
@@ -274,7 +286,7 @@ class Transformer_InPlace(Transformer):
         return self._transform_tree(tree)
 
 
-class Transformer_NonRecursive(Transformer):
+class Transformer_NonRecursive(Transformer[_Leaf_T, _Return_T]):
     """Same as Transformer but non-recursive.
 
     Like Transformer, it doesn't change the original tree.
@@ -321,7 +333,7 @@ class Transformer_NonRecursive(Transformer):
         return cast(_Return_T, result)
 
 
-class Transformer_InPlaceRecursive(Transformer):
+class Transformer_InPlaceRecursive(Transformer[_Leaf_T, _Return_T]):
     "Same as Transformer, recursive, but changes the tree in-place instead of returning new instances"
     def _transform_tree(self, tree):
         tree.children = list(self._transform_children(tree.children))
@@ -462,8 +474,7 @@ class _VArgsWrapper:
     def __init__(self, func: Callable, visit_wrapper: Callable[[Callable, str, list, Any], Any]):
         if isinstance(func, _VArgsWrapper):
             func = func.base_func
-        # https://github.com/python/mypy/issues/708
-        self.base_func = func  # type: ignore[assignment]
+        self.base_func = func
         self.visit_wrapper = visit_wrapper
         update_wrapper(self, func)
 
@@ -472,19 +483,21 @@ class _VArgsWrapper:
 
     def __get__(self, instance, owner=None):
         try:
-            g = self.base_func.__get__
+            # Use the __get__ attribute of the type instead of the instance
+            # to fully mirror the behavior of getattr
+            g = type(self.base_func).__get__
         except AttributeError:
             return self
         else:
-            return _VArgsWrapper(g(instance, owner), self.visit_wrapper)
+            return _VArgsWrapper(g(self.base_func, instance, owner), self.visit_wrapper)
 
     def __set_name__(self, owner, name):
         try:
-            f = self.base_func.__set_name__
+            f = type(self.base_func).__set_name__
         except AttributeError:
             return
         else:
-            f(owner, name)
+            f(self.base_func, owner, name)
 
 
 def _vargs_inline(f, _data, children, _meta):
@@ -498,19 +511,19 @@ def _vargs_tree(f, data, children, meta):
 
 
 def v_args(inline: bool = False, meta: bool = False, tree: bool = False, wrapper: Optional[Callable] = None) -> Callable[[_DECORATED], _DECORATED]:
-    """A convenience decorator factory for modifying the behavior of user-supplied visitor methods.
+    """A convenience decorator factory for modifying the behavior of user-supplied callback methods of ``Transformer`` classes.
 
-    By default, callback methods of transformers/visitors accept one argument - a list of the node's children.
+    By default, transformer callback methods accept one argument - a list of the node's children.
 
-    ``v_args`` can modify this behavior. When used on a transformer/visitor class definition,
-    it applies to all the callback methods inside it.
+    ``v_args`` can modify this behavior. When used on a ``Transformer`` class definition, it applies to
+    all the callback methods inside it.
 
     ``v_args`` can be applied to a single method, or to an entire class. When applied to both,
     the options given to the method take precedence.
 
     Parameters:
         inline (bool, optional): Children are provided as ``*args`` instead of a list argument (not recommended for very long lists).
-        meta (bool, optional): Provides two arguments: ``children`` and ``meta`` (instead of just the first)
+        meta (bool, optional): Provides two arguments: ``meta`` and ``children`` (instead of just the latter); ``meta`` isn't available for transformers supplied to Lark using the ``transformer`` parameter (aka internal transformers).
         tree (bool, optional): Provides the entire tree as the argument, instead of the children.
         wrapper (function, optional): Provide a function to decorate all methods.
 
@@ -522,6 +535,12 @@ def v_args(inline: bool = False, meta: bool = False, tree: bool = False, wrapper
                 def add(self, left, right):
                     return left + right
 
+                @v_args(meta=True)
+                def mul(self, meta, children):
+                    logger.info(f'mul at line {meta.line}')
+                    left, right = children
+                    return left * right
+
 
             class ReverseNotation(Transformer_InPlace):
                 @v_args(tree=True)
